/*
 * mod_auth_imap: authentication via IMAP server
 * Version: 0.1a - First ALPHA release.
 *
 * brillat-apachemodule@mainsheet.org (Ben Brillat), based on mod_auth_pop by:
 * awl@verat.net (Milos Prodanovic)
 */

/*
 *
 */


#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_protocol.h"
#include "http_conf_globals.h"

#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <netdb.h>
#include <syslog.h>
 
#define _OK 1
 
int Sock;

int tcp_gets(int s, char *res,int len) {
    char c;
    int cur=0,rc;
 
  memset(res,0,len+1);
  
    while( (rc = read(s, &c, 1))!=EOF) {
	if(cur<len)
	res[cur]=c;
        cur++;
        if(c=='\n')
        	break;     
             }
return 0;
}

void clean_up(int s) { 
	close(s);
}

int tcp_puts(int s, char *sttr) {
int rr,len;
char line[512];

memset(line,0,512);
len=strlen(sttr);
if(len>510) len=510;
strncpy(line,sttr,len);
rr=write(s,line,strlen(line));
return rr;
}


int imap_tcp_open(char *hostname, int port) {
    struct hostent *HOST;
    struct sockaddr_in sai;
    int s;


    if((HOST = gethostbyname(hostname)) == (struct hostent *)NULL) {
        ap_log_error(APLOG_MARK,APLOG_WARNING|APLOG_NOERRNO,NULL,"mod_auth_imap: Hostname unknown.");
        return !_OK;
    }
    memset((char *)&sai, '\0', sizeof(sai));
    memcpy((char *)&sai.sin_addr, (char *)HOST -> h_addr,
           HOST -> h_length);
    sai.sin_family = HOST -> h_addrtype;
    sai.sin_port = htons(port);
 
    if((s = socket(HOST -> h_addrtype, SOCK_STREAM, 0)) == -1) {
        ap_log_error(APLOG_MARK,APLOG_WARNING|APLOG_NOERRNO,NULL,"mod_auth_imap: socket problem");
        clean_up(s);
        return !_OK;
    }
 
    if(connect(s, (struct sockaddr *)&sai, sizeof(sai)) == -1) {
        ap_log_error(APLOG_MARK,APLOG_WARNING|APLOG_NOERRNO,NULL,"mod_auth_imap: connect() problem");
        clean_up(s);
 	return !_OK;
   }
    return s;
}
 
int imap_auth_a3(char *host,char *username,char *pass,char *cport) {
char result[512],buf[351];
int ret=0;
int port;

	port=atoi(cport);

if( strlen(username)>100 || strlen(pass)>100 )
{
	ap_log_error(APLOG_MARK,APLOG_WARNING|APLOG_NOERRNO,NULL,"username/password too long for mod_auth_imap");
	printf("Ouch - u/p too long!\n");
	return !_OK;
}
else
{
	
	Sock=imap_tcp_open(host,port);
	if(!Sock)
	{
		return !_OK;
	}


	tcp_gets(Sock,result,350);


	memset(buf,0,350);
	sprintf(buf,"A001 LOGIN %s %s\n", username, pass);


	tcp_puts(Sock,buf);
	tcp_gets(Sock,result,350);

	
	if(!strncmp(result,"A001 OK",7))
		{
		ret=_OK;
		}
	else
	{
		ap_log_error(APLOG_MARK,APLOG_WARNING|APLOG_NOERRNO,NULL,"mod_auth_imap: bad login for user %s.", username);
		ret=!_OK;
	}
	memset(buf,0,350);

	sprintf(buf,"A002 LOGOUT\n");
	tcp_puts(Sock,buf);
	tcp_gets(Sock,result,350);

}

clean_up(Sock);


return ret;
}
 

typedef struct auth_config_struct {
    char *imap_server;
    char *imap_port;
    int   imap_authoritative;
    int   enabled;
} imap_config_rec;

typedef struct {
  char *name, *pw;
} auth_imap_userinfo;


static void *create_imap_dir_config (pool *p, char *d)
{
    imap_config_rec *sec =
	(imap_config_rec *) ap_palloc (p, sizeof(imap_config_rec));
    sec->imap_server = "127.0.0.1";     /* localhost is server by default */
    sec->imap_port = "143"; 	/* port is 143 by default */
    sec->imap_authoritative = 1; /* keep the fortress secure by default */
    sec->enabled = 0; /* disable this authentication by defualt */
    return sec;
}

static const char *set_imap_slot (cmd_parms *cmd, char *struct_ptr, char *arg)
{
   int offset = (int)cmd->info;
   *(char **)(struct_ptr + offset) = ap_pstrdup (cmd->pool, arg);
   return NULL;
}

static const char *set_flag_slot (cmd_parms *cmd, imap_config_rec *sec, char *arg)
{  /*warning - the next line generates warnings.  it's ok. */
   sec->imap_authoritative = (unsigned char) arg;
   return NULL;
}

static
char* auth_enable(cmd_parms *cmd, imap_config_rec *config, int arg)
{
  config->enabled = arg;
  return NULL;
}

static command_rec imap_cmds[] = {
{ "Auth_IMAP_Server", set_imap_slot,
  (void*)XtOffsetOf(imap_config_rec,imap_server), OR_AUTHCFG, TAKE1,
     "IMAP server . " },
{ "Auth_IMAP_Port", set_imap_slot,
  (void*)XtOffsetOf(imap_config_rec,imap_port), OR_AUTHCFG, TAKE1,
     "IMAP port . " },
{ "Auth_IMAP_Authoritative", set_flag_slot,
  (void*)XtOffsetOf(imap_config_rec,imap_authoritative),
    OR_AUTHCFG, FLAG, 
   "Set to 'no' to allow access control to be passed along to lower modules if the UserID is not known to this module" },
{ "Auth_IMAP_Enabled", (const char*(*)())auth_enable, NULL, OR_AUTHCFG, FLAG,
    "on|off - determines if IMAP authentication is enabled; default is off" },
{ NULL }
};

module auth_imap_module;


static int imap_authenticate_basic_user (request_rec *r)
{
    imap_config_rec *sec =
      (imap_config_rec *)ap_get_module_config (r->per_dir_config, &auth_imap_module);
    conn_rec *c = r->connection;
    int res,i;
    char *server = sec->imap_server;
    char *port = sec->imap_port;
    auth_imap_userinfo userinfo = { NULL, NULL }; 


/* is this module disabled? */
  if (!sec->enabled)
    return DECLINED;
 
    if ((res = ap_get_basic_auth_pw (r, (const char**)&(userinfo.pw)))) return res;
   

  /* this is only set after get_basic_auth_pw was called */
  userinfo.name = r->connection->user;
 
    if(!sec->imap_server)  
        return DECLINED;

    /*
     * Check in User-Name
     */
    /* send IMAP authen, server port, username, sent_pw */
  
    i=imap_auth_a3(server,userinfo.name,userinfo.pw,port);    

    if (i==1) return (OK);
    else return AUTH_REQUIRED;
}
    
/* Checking ID */
    
static int imap_check_user_access (request_rec *r) {
    imap_config_rec *sec =
     (imap_config_rec *)ap_get_module_config (r->per_dir_config, &auth_imap_module);
    char *user = r->connection->user;
    int m = r->method_number;
    int method_restricted = 0;
    register int x;
    const char *t, *w;
    array_header *reqs_arr = ap_requires (r);
    require_line *reqs;

    if (!reqs_arr)
        return (OK);
    reqs = (require_line *)reqs_arr->elts;
  
    for(x=0; x < reqs_arr->nelts; x++) {
      
	if (! (reqs[x].method_mask & (1 << m))) continue;
	
	method_restricted = 1;

        t = reqs[x].requirement;
        w = ap_getword(r->pool, &t, ' ');
        if(!strcmp(w,"valid-user"))
            return OK;
        if(!strcmp(w,"user")) {
            while(t[0]) {
                w = ap_getword_conf (r->pool, &t);
                if(!strcmp(user,w))
                    return OK;
            }
        }
	else if(!strcmp(w,"group"))
	        return DECLINED;
    }
    
    if (!method_restricted)
      return OK;

    if (!(sec->imap_authoritative))
      return DECLINED;

    ap_note_basic_auth_failure (r);
    return AUTH_REQUIRED;
}

module auth_imap_module = {
   STANDARD_MODULE_STUFF,
   NULL,                        /* initializer */
   create_imap_dir_config,       /* dir config creater */
   NULL,			/* dir merger --- default is to override */
   NULL,			/* server config */
   NULL,			/* merge server config */
   imap_cmds,                    /* command table */
   NULL,			/* handlers */
   NULL,			/* filename translation */
   imap_authenticate_basic_user, /* check_user_id */
   imap_check_user_access,       /* check auth */
   NULL,			/* check access */
   NULL,			/* type_checker */
   NULL,			/* fixups */
   NULL,			/* logger */
   NULL				/* header parser */
};




